TIMEOUT 1000
1 SEND '$*\n'
2 THROW
3 IN '\n' 100
4 IF '30' 10
5 IF '10' 20
6 IF '0000 DATA END WITH \'.\'' 30
7 GOTO 2

10 SKIP 5
11 COPY
13 OUTTAG 3
14 OUTSHIFT
15 GOTO 2

20 SKIP 2
21 RVATOI
22 THROW
23 IN '$ ' 100
24 IF '30' 40
25 EXITRV

30 THROW
31 IN '\n' 100
32 IF '.\n' 35
33 COPY
34 GOTO 30
35 OUTTAG 1
36 OUTSHIFT
37 GOTO 2

40 SKIP 5
41 COPY
42 OUTTAG 3
43 OUTSHIFT
44 SKIP '\n'
45 GOTO 24

100 EXIT -1

